/*
 *  display.c --- handle the curses windows for the oms com program.
 *
 *    Copyright(c) 2003 Oregon Micro Systems, Inc.  All rights reserved.
 */
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <asm/types.h>   // need cpu-specific u32
#include <fcntl.h>
#include <sys/time.h>
#include <sys/select.h>
#include <sys/ioctl.h>
#include <time.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <curses.h>
#include <malloc.h>
#include <unistd.h>

#include "list.h"          // non-kernel version of kernel header file

extern FILE *transcript_fp;
extern char *ourname;      

// all of these are exported from *com.c
extern void  delete_nonprinting(char *buf);
extern void  delete_comment(char *buf);
extern void  delete_trailing_whitespace(char *buf);
extern int   line_is_empty(char *buf);
extern void  write_line_to_controller(char *buf);
extern int   check_output_from_controller(char *buf);
extern void  get_driver_version_string(char *buf);
extern void  capture_position_data(char *buf);

// some private function prototypes.  (Sadly, some functions here
//    are merely declared before use, not prototypes.  Fix when convenient.)
static WINDOW *cardwin, *interactwin;
static int     cardwin_lines, cardwin_cols;
static int     interactwin_lines, interactwin_cols;
static FILE   *script_file_fp;           // also used as boolean
static int     reading_script_filename;  // boolean


// we export two functions
void   run_display(void);
void   quit_curses(void) { endwin(); }

static char *filename_prompt = "enter file name:  ";


// Home-build scrolling.  I've never been very happy with curses
//   scrolling.  It is tricky to figure out how to do, and sufficiently
//   non-intuitive to remember once figured out.  Plus, I've seen
//   several curses programs that use scrolling that is broken (at
//   least on some terminal types).  Back in the days when we used
//   serial-port terminals running at 9600 baud, being able to use
//   the hardware scrolling of the terminal was a BIG win.  For xterms
//   or VGA consoles, it is not cost-prohibitive to simply repaint the
//   entire window with each new line added.  Accordingly, here's a
//   few data structures and functions that let me keep track of
//   what is displayed and handle scrolling by erase/redraw.
#define MAX_LINE_LEN 2048

struct w_line {
        struct list_head listhead;
        char   line[MAX_LINE_LEN];
};

static LIST_HEAD(freelines);

static struct w_line *
alloc_w_line(void)
{
        struct w_line    *new;
        struct list_head *pos;

        if (list_empty(&freelines)) {
                if ((new = malloc(sizeof(struct w_line))) == NULL) {
                        quit_curses();
                        fprintf(stderr, "%s:  out of memory\n", ourname);
                        exit(1);
                }
        } else {
                pos = freelines.next;
                new = list_entry(pos, struct w_line, listhead);
                list_del(pos);
        }
        INIT_LIST_HEAD(&new->listhead);
        return new;
}

static void
free_w_line(struct w_line *wl)
{
        struct list_head *pos = &wl->listhead;
        
        list_add(pos, &freelines);
}

struct wcontents {
        WINDOW          *w;
        int              max_linecount;
	int              win_cols;   // number of COLS in the window
        int              curr_linecount;
        struct list_head lines;
        int              have_initialized;
};

static struct wcontents interact_contents, card_contents;

static void
init_wcontents_struct(struct wcontents *wc)
{
        if (wc->have_initialized)
                return;
        wc->have_initialized = 1;
        INIT_LIST_HEAD(&wc->lines);
        wc->curr_linecount = 0;
        if (wc == &interact_contents) {
                wc->w = interactwin;
                wc->max_linecount = interactwin_lines - 2;
		wc->win_cols = interactwin_cols - 3;
        } else if (wc == &card_contents) {
                wc->w = cardwin;
                wc->max_linecount = cardwin_lines - 2;
		wc->win_cols = cardwin_cols - 3;
        } else {
                quit_curses();
                fprintf(stderr, "%s:  programmer error %s:%d\n",
                        ourname, __FILE__, __LINE__);
                exit(1);
        }
        // move the cursor to (1,1) so that it looks right before
        // anything is put in.
        wmove(wc->w, 1, 1);
        wrefresh(wc->w);
}

static void
redisplay_wcontents(struct wcontents *wc)
{
        struct w_line    *cur;
        struct list_head *pos;
        int               i;

        if (!wc->have_initialized)
                init_wcontents_struct(wc);

        if (list_empty(&wc->lines)) {
                return;
        }

        i = 1;
        werase(wc->w);
        list_for_each(pos, &wc->lines) {
                cur = list_entry(pos, struct w_line, listhead);
                mvwprintw(wc->w, i, 1, "%s", cur->line);
                i++;
        }
        wborder(wc->w, 0,0, 0,0, 0,0, 0,0);
        wrefresh(wc->w);
        // if window is not interactwin, refresh that window to
        //    make the visible cursor appear there.
        if (wc->w != interactwin)
                wrefresh(interactwin);
}

static void
add_line_to_wcontents(struct wcontents *wc, char *buffer)
{
        struct w_line    *new, *old;
        struct list_head *pos;
        int used, len = strlen(buffer);
        char* buf = buffer;
        char* cr;

        if (!wc->have_initialized)
                init_wcontents_struct(wc);
        // don't overfill a curses window line.  wrap
        if ( len == 0 )
        {
          new = alloc_w_line();
          strncpy(new->line, buf, wc->win_cols);
          list_add_tail(&new->listhead, &wc->lines);
          ++(wc->curr_linecount);
          if (wc->curr_linecount > wc->max_linecount) 
          {
                  // toss the oldest
                  pos = wc->lines.next;
                  list_del(pos);
                  old = list_entry(pos, struct w_line, listhead);
                  free_w_line(old);
                  --(wc->curr_linecount);
          }
        }
        else
        {        
          while ( len > 0 )
          {
            new = alloc_w_line();
            strncpy(new->line, buf, wc->win_cols);
            if ((cr = strchr(new->line, '\n')) != NULL)
                    cr[0] = 0;
            used = strlen( new->line );
            if ( cr != NULL )
                    used++;
            list_add_tail(&new->listhead, &wc->lines);
            ++(wc->curr_linecount);
            if (wc->curr_linecount > wc->max_linecount) 
            {
                    // toss the oldest
                    pos = wc->lines.next;
                    list_del(pos);
                    old = list_entry(pos, struct w_line, listhead);
                    free_w_line(old);
                    --(wc->curr_linecount);
            }
            len -= (used);
            buf += (used);
          }
        }
        redisplay_wcontents(wc);
}

static void
add_char_to_wcontents(struct wcontents *wc, char c)
{
        struct list_head *pos;
        struct w_line    *cur;
        int               len;

        if (!wc->have_initialized)
                init_wcontents_struct(wc);
        if (list_empty(&wc->lines)) {
                // add an empty line
                add_line_to_wcontents(wc, "");
        }
        pos = wc->lines.prev;
        cur = list_entry(pos, struct w_line, listhead);
        len = strlen(cur->line);
        if (len < wc->win_cols) {
		// quick 'n' dirty:  just DROP byte if it won't fit
                cur->line[len] = c;
                cur->line[len + 1] = 0;  // always keep string nul-delimited
                redisplay_wcontents(wc);
        }
}

static void
get_last_wline(struct wcontents *wc, char *buf)
{
        struct list_head *pos;
        struct w_line    *cur;
 
        if (!wc->have_initialized)
                init_wcontents_struct(wc);
        if (list_empty(&wc->lines)) {
                *buf = 0;
                return;
        }
        pos = wc->lines.prev;
        cur = list_entry(pos, struct w_line, listhead);
        strcpy(buf, cur->line);
}

// delete_char_from_wcontents() --- backspace.  Don't allow deletion
//   beyond beginning of line.
//
//  Special case:  if reading_script_filename is TRUE, then don't
//    allow deletion into the prompt that's in the line.
static void
delete_char_from_wcontents(struct wcontents *wc)
{
        struct list_head *pos;
        struct w_line    *cur;
        int               len;

        if (!wc->have_initialized)
                init_wcontents_struct(wc);
        if (list_empty(&wc->lines))
                return;
        pos = wc->lines.prev;
        cur = list_entry(pos, struct w_line, listhead);
        len = strlen(cur->line);
        if (len == 0)
                return;
	if (reading_script_filename) {
		if (len <= strlen(filename_prompt))
			return;
	}
        cur->line[len - 1] = 0;
        redisplay_wcontents(wc);
}

// empty_the_last_line() --- make the last line in a wcontents struct
//   empty.  Though it should never happen, we make sure that there
//   IS a last line.
static void
empty_the_last_line(struct wcontents *wc)
{
	struct list_head *pos;
	struct w_line    *cur;

	if (!wc->have_initialized)
		init_wcontents_struct(wc);
	if (list_empty(&wc->lines)) {
		add_line_to_wcontents(wc, "");  // put in an empty line
		return;
	}
	pos = wc->lines.prev;
	cur = list_entry(pos, struct w_line, listhead);
	cur->line[0] = 0;
	redisplay_wcontents(wc);
}

// add_string_to_last_line() --  strcat a string to the end of the last line
//   in a wcontents struct.
static void
add_string_to_last_line(struct wcontents *wc, char *str)
{
	struct list_head *pos;
	struct w_line    *cur;
	int               len = strlen(str);

	if (!wc->have_initialized)
		init_wcontents_struct(wc);
	if (list_empty(&wc->lines)) {
		add_string_to_last_line(wc, str);
		return;
	}
	pos = wc->lines.prev;
	cur = list_entry(pos, struct w_line, listhead);
	// if total length exceeds curses window width, toss some bytes
	while ((strlen(cur->line) + len) >= wc->win_cols) {
		len--;
		str[len] = 0;
	}
	strcat(cur->line, str);
	redisplay_wcontents(wc);
}

static int
user_input_char_avail(char *newchar)
{
        int   c = getch();

        if (c == ERR)
                return 0;
        if (c == KEY_ENTER || c == '\r' || c == '\n') {
                *newchar = '\r';
                return 1;
        }
        if (c == KEY_DC || c == KEY_BACKSPACE || c == '\b') {
                *newchar = '\b';
                return 1;
        }
	if (c == KEY_F(1)) {    // F1
		*newchar = 3;   // translate as ^C
		return 1;
	}
	if (c == KEY_F(5)) {		// F5
		*newchar = 2;		// translate as ^b
		return 1;
	}
        if (c == 3) {          //  real ^C (should never happen -- cbreak)
                *newchar = c;
                return 1;
        }
        if (c >= ' ' && c <= '~') {
                *newchar = c;
                return 1;
        }
        return 0;
}

// open_script_file() --- the last line from the interact_contents struct
//   has been fetched into a buffer, which we may use.  See if the
//   buffer contains a readable file name, and if so, open it and log
//   the fact that we're now processing a script.  If the file name in
//   the buffer is empty, empty out the last line in interact_contents and
//   return.  If we fail to open the name, log an error message and return.
static void
open_script_file(char *buf)
{
        char  logbuf[MAX_LINE_LEN];

        // first, point beyond our prompt.
        buf += strlen(filename_prompt);

        // skip past any leading white space
        while (*buf && (*buf == ' ' || *buf == '\t'))
                buf++;
        // delete any trailing white space
        delete_trailing_whitespace(buf);

        // as it turns out, all paths through this function cause us
        // to empty the last line in interact_contents.  Do it now.
        empty_the_last_line(&interact_contents);

        // if there's nothing left, just return (the user might have
        //  backspaced to the prompt and hit <enter>)
        if (strlen(buf) == 0) {
                return;
        }
        if ((script_file_fp = fopen(buf, "r")) == NULL) {
                sprintf(logbuf, " -->> can't read \"%s\":  %s",
                        buf, strerror(errno));
                add_string_to_last_line(&interact_contents, logbuf);
                if (transcript_fp) {
                        fprintf(transcript_fp, "%s\n", logbuf);
                        fflush(transcript_fp);
                }
                // make window ready for new user input
                add_line_to_wcontents(&interact_contents, "");
                return;
        }
        // we successfully opened the script file
        sprintf(logbuf, "  **** Reading from file \"%s\"", buf);
        add_string_to_last_line(&interact_contents, logbuf);
        if (transcript_fp) {
                fprintf(transcript_fp, "%s\n", logbuf);
                fflush(transcript_fp);
        }
        // make window ready for first line of the script.
        add_line_to_wcontents(&interact_contents, "");
}

        
// process_new_char() --- process a new char typed in by the
//   user.  There are three modes:
//     1.  normal INTERACTIVE_MODE processing.
//     2.  user input typed while processing a script file.
//         (we ignore everything but F1, which is translated to us as ^C)
//     3.  receiving a script file name after prompting for it.
static void
process_new_char(char newchar)
{
        char buf[MAX_LINE_LEN];

	// handle user request for position data
	if (newchar == 2) {
		capture_position_data(&buf[0]);
    add_line_to_wcontents(&card_contents, &buf[0]);
		return;
	}

	// if we're procssing a script file, ignore all user input
	//  except F1 (translated to us as ^C).
	if (script_file_fp) {
		if (newchar == 3) {
			// aborted
			fclose(script_file_fp);
			script_file_fp = NULL;
			empty_the_last_line(&interact_contents);
			add_string_to_last_line(&interact_contents,
						" *** SCRIPT ABORTED ***");
			add_line_to_wcontents(&interact_contents, "");
			if (transcript_fp) {
				fprintf(transcript_fp, 
					" *** SCRIPT ABORTED ***");
				fflush(transcript_fp);
			}
		}
		// discard all other typing during script file processing
		return;
	}
	// if the user is entering a file name, we process chars nearly
	//  the same as in normal mode, but of course we don't send the
	//  filename to the controller.
	if (reading_script_filename) {
		if (newchar == '\n' || newchar == '\r') {
			get_last_wline(&interact_contents, buf);
			open_script_file(buf);
                        reading_script_filename = 0;
		} else if (newchar == 3) {
			// user aborts file name entry
			//  remove the file-entry prompt and any partial
			//  name typed.
			empty_the_last_line(&interact_contents);
			reading_script_filename = 0;
		} else if (newchar == '\b') {
			delete_char_from_wcontents(&interact_contents);
		} else if (newchar >= ' ') {
			add_char_to_wcontents(&interact_contents, newchar);
		}
		return;
	}

	// get here for normal INTERACTIVE_MODE procesing of typed chars
        if (newchar == '\n' || newchar == '\r') {
                get_last_wline(&interact_contents, buf);
                write_line_to_controller(buf);
                add_line_to_wcontents(&interact_contents, "");
	} else if (newchar == 3) {
		// user wants to run a command script
		empty_the_last_line(&interact_contents);
		add_string_to_last_line(&interact_contents, filename_prompt);
		reading_script_filename = 1;
        } else if (newchar == '\b') {
                delete_char_from_wcontents(&interact_contents);
        } else if (newchar >= ' ') {
                add_char_to_wcontents(&interact_contents, newchar);
        }
}

static void
process_controller_output(char *buf)
{
        add_line_to_wcontents(&card_contents, buf);
}

static void
read_next_line_from_script_file(void)
{
        char  buf[MAX_LINE_LEN];
        char *fgets_rv;

        while (1) {
                errno = 0;
                fgets_rv = fgets(buf, sizeof(buf), script_file_fp);
                if (fgets_rv != NULL)
                        break;
                if (errno != EINTR)
                        break;
        }
        if (fgets_rv == NULL) {
                // EOF on script file
                fclose(script_file_fp);
                script_file_fp = NULL;  // for use as boolean
                add_string_to_last_line(&interact_contents,
                                        " ***** script file EOF *****");
                if (transcript_fp) {
                        fprintf(transcript_fp, "%s\n",
                                " ***** script file EOF *****");
                        fflush(transcript_fp);
                }
                // make interact_contents ready for user input
                add_line_to_wcontents(&interact_contents, "");
                return;
        }
        // process the line from the file
        delete_nonprinting(buf);
        delete_comment(buf);
        delete_trailing_whitespace(buf);
        if (!line_is_empty(buf)) {
                add_string_to_last_line(&interact_contents, buf);
                write_line_to_controller(buf);
                add_line_to_wcontents(&interact_contents, "");
        }
}


// run_display() --- the main loop of the program.
//   This is called once the controller is opened and put
//   in INTERACTIVE_MODE.  We initialize the curses display and
//   poll both the user-input queue and the controller.
//
//  We never return.  Normal program exit is via SIGINT, which the
//  curses library catches and uses to clean up.
void
run_display(void)
{
	char    buf[MAX_LINE_LEN];
	char    newchar;

	// yawn -- normal curses setup stuff
	initscr(); cbreak(); noecho(); nonl();
	intrflush(stdscr, FALSE); keypad(stdscr, TRUE);

	// halfdelay() is important.  It allows pretty responsive
	//   control without hogging all CPU cycles.
	halfdelay(1);

	// create our two windows.  Make the 'cardwin' window (showing
	//  output from the controller) about 2/3 of the total screen.
	cardwin_lines = (LINES / 3) * 2;
	cardwin_cols  = COLS - 1;
	if ((cardwin = newwin(cardwin_lines, cardwin_cols, 1, 1)) == NULL) 
	{
		quit_curses();
		exit(1);
	}
	if (wborder(cardwin, 0,0, 0,0, 0,0, 0,0) == ERR) 
	{
		quit_curses();
		fprintf(stderr, "wborder of cardwin failed\n");
		exit(2);
	}
	interactwin_lines = LINES / 3;
	interactwin_cols  = COLS - 1;
	while (1) 
	{
		if ((interactwin = newwin(interactwin_lines,
															interactwin_cols,
															cardwin_lines + 1, 1)) != NULL) 
		{
			break;
		}
		if (--interactwin_lines <= 3) 
		{
			quit_curses();
			exit(3);
		}
	}
	if (wborder(interactwin, 0,0, 0,0, 0,0, 0,0) == ERR) 
	{
		quit_curses();
		fprintf(stderr, "wborder of interactwin failed\n");
		exit(2);
	}
	wrefresh(cardwin);
	wrefresh(interactwin);

	// KLUDGE ALERT
	//  for some reason, curses likes to erase the interactive
	//  window until a byte is processed through getch().
	//  Okay, we'll play along.  Place an innocuous message in
	//  the window, except the last character, which is put into
	//  curses's input mechanism with ungetch().  Read it back
	//  out and Bob's your uncle.
	ungetch('x');
	while (!user_input_char_avail(&newchar));
	add_line_to_wcontents(&interact_contents, "User Input Window");
	add_line_to_wcontents(&interact_contents, "");
	// end kludge alert (my uncle is Bob)

	// put the device driver version string in the card_contents window
	get_driver_version_string(buf);
	add_line_to_wcontents(&card_contents, buf);

	while (1) 
	{
		if (user_input_char_avail(&newchar))
			process_new_char(newchar);
		if (check_output_from_controller(buf)) 
		{
			process_controller_output(buf);
		}
		if (script_file_fp) 
		{
			read_next_line_from_script_file();
		}
	}
	// not reached
}

