/*
 *  oms-maxpdrv.c --- Linux device driver for Oregon Micro Systems
 *    MAXp motion controller, hardware-specific code.
 *
 *  Copyright(c) 2003 Oregon Micro Systems, Inc.  All rights reserved.
 */

char version_string[] = "Oregon Micro Systems MAXp Linux Driver v2.0";
char* me = "MAXp";

#ifndef __KERNEL__
#define __KERNEL__
#endif
#ifndef MODULE
#define MODULE
#endif

#include <linux/config.h>
#include <linux/version.h>

#ifdef CONFIG_MODVERSIONS
#ifndef MODVERSIONS
#define MODVERSIONS
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,5,40)
#include <linux/modversions.h>
#endif
#endif

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/interrupt.h>
#include <linux/timer.h>
#include <linux/pci.h>
#include <linux/spinlock.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/io.h>
#include <linux/ioctl.h>
#include <linux/kmod.h>
#include <asm/atomic.h>
#include <asm/uaccess.h>
#include <asm/byteorder.h>
#include <linux/byteorder/generic.h>

#include "oms-core.h"      // driver-private header
#include "oms-maxpdrv.h"   // driver-private card-specific header

MODULE_DESCRIPTION("Driver for OMS MAXp motion control");
char* node_device = "maxp";

//****************************************************************************
// PCI BUS VENDOR AND DEVICE IDS FOR BUS INTERFACE AND OMS DEVICES

// The bus interface vendor/device IDs will be in system vendor ID
//  registers of PCI configuration space
#define PCI_VENDOR_ID_BUS	    	0x1057
#define PCI_DEVICE_ID_BUS				0x0006
// The OMS vendor/device IDs will be in the subsystem vendor ID
//   registers of PCI configuration space
#define PCI_VENDOR_ID_OMS				0x160c
#define PCI_DEVICE_NAME					"maxp"

int PCI_DEVICE_ID_OMS_BASE = 0x0010;

// pci_device_id --- specify to PCI driver the set of vendor/device IDs
//     that we know about.  The PCI driver will call our "probe" routine
//     for each device matching one of our descriptions.
struct pci_device_id oms_pci_tbl[] __devinitdata = 
{
	{ PCI_VENDOR_ID_BUS,
	  PCI_DEVICE_ID_BUS,
	  PCI_VENDOR_ID_OMS,
	  PCI_ANY_ID    // we will accept anything >= PCI_OMS_DEVICE_ID_BASE
	},
	{ 0 }
};


// pci_driver structure --- register with the PCI driver.  It will call
//   our "probe" routine for each PCI device matching our table.
struct pci_driver oms_driver = 
{
	.name     = PCI_DEVICE_NAME,
	.id_table = oms_pci_tbl,
	.probe    = oms_install_one,
	.remove   = __devexit_p(oms_remove_one),
};



//****************************************************************************
// externs from oms-core.c
extern struct list_head oms_cardlist;
extern unsigned long flags;
// Module parameters
extern int DEBUGLEVEL;
extern int MAJORNUMBER;
extern int ONE_CPU;
extern int MAX_CARDS;
extern char* NODE_CREATION_SCRIPT;
#if OMS_DEBUG_SPINLOCKS
	extern atomic_t   oms_islocked;
#endif


//****************************************************************************
// Private functions
void _oms_irq_handle_ibf( struct oms_card* card );
void _oms_irq_handle_pos_data( struct oms_card* card );



//****************************************************************************
// _oms_init_memory ---- called by oms_install_one to probe for
//   card-specific memory regions and allocate them
//
//  We don't know the index numbers of the 3 BARs we need.  One
//    is 4K, one is 64K and one is 512 bytes.  Search for these.
int _oms_init_memory( struct pci_dev* pdev,
										  struct oms_card* card )
{
	int             	i;
	int             	bars_found = 0;
	unsigned long   	start, len, siz;

	card->hardware = (void*)kmalloc( sizeof(struct hw_struct), GFP_ATOMIC );
	memset(card->hardware, 0, sizeof(struct hw_struct));
	
	// scan the BARs looking for MMIO regions of 4k, 64k and 512k
	for ( i = DEVICE_COUNT_RESOURCE; bars_found < 3 && i >= 0; i-- )
	{
		start = pci_resource_start( pdev, i );
		len = pci_resource_len( pdev, i );

		if ( start == 0 || len == 0 )
			continue;
		if ( (pci_resource_flags( pdev, i ) & IORESOURCE_MEM) == 0 )
			continue;

		siz = 4 * 1024;  // look for 4k register space
		if ( len == siz )
		{
			if ( !request_mem_region( start, siz, me ) )
			{
				printk(KERN_ERR "%s:  mem region collision at %p\n", me, (void*) start);
					return 1;
			} 
			// we store the base of the mapped registers
			//  as an unsigned long, as we never dereference
			//  it, but need the value for address arithmetic
			//  to derive the pointers we actually use.
			card_hw->regbase_addr = (unsigned long)ioremap( start, siz );

			if ( !card_hw->regbase_addr )
			{
				printk(KERN_ERR "%s:  ioremap failed\n", me);
				return 1;
			}
			card_hw->regbase_physaddr = start;

			// set pointers to registers
			card_hw->obisr = (u32*)(card_hw->regbase_addr + REG_OBISR);
			card_hw->obimr = (u32*)(card_hw->regbase_addr + REG_OBIMR);
			card_hw->obmr0 = (u32*)(card_hw->regbase_addr + REG_OBMR0);
			card_hw->obmr1 = (u32*)(card_hw->regbase_addr + REG_OBMR1);
			card_hw->obdr  = (u32*)(card_hw->regbase_addr + REG_OBDR);

			bars_found++;
			// continue 'for' loop
			continue;
		}
                        
		siz = 64 * 1024;   // look for shared memory
		if ( len == siz )
		{
			if ( !request_mem_region( start, siz, me ) )
			{
				printk(KERN_ERR "%s:  mem region collision at %p\n", me, (void*) start);
				return 1;
			} 
			card_hw->shared64k = ioremap( start, siz );
			if ( !card_hw->shared64k )
			{
				printk(KERN_ERR "%s:  ioremap failed\n", me);
				return 1;
			}
			card_hw->shared64k_physaddr = start;

			// set pointers
			card_hw->firmware  = (u32*)&card_hw->shared64k[MEM_FIRMWARE];
			card_hw->limit     = (u32*)&card_hw->shared64k[MEM_RT_LIMIT];
			card_hw->home      = (u32*)&card_hw->shared64k[MEM_RT_HOME];
			card_hw->semaphore = (u32*)&card_hw->shared64k[MEM_OM_SEMAPHORE];
			card_hw->mailbox   = (u32*)&card_hw->shared64k[MEM_CAP_RQ_MBOX];
		
			// Set up pointers to command and response buffers
			card_hw->command_buf  = (char*)(&card_hw->shared64k[MEM_CMD_BUFFER]);
			card_hw->command_ip   = (u32*)&card_hw->shared64k[MEM_CMD_INSERT];
			card_hw->command_pp   = (u32*)&card_hw->shared64k[MEM_CMD_PROCESS];
			card_hw->response_buf = (char*)(&card_hw->shared64k[MEM_RESP_BUFFER]);
			card_hw->response_ip  = (u32*)&card_hw->shared64k[MEM_RESP_INSERT];
			card_hw->response_pp  = (u32*)&card_hw->shared64k[MEM_RESP_PROCESS];

			bars_found++;
			// continue 'for' loop
			continue;
		}

		siz = 512 * 1024;   // look for large shared memory
		if ( len == siz )
		{
			if ( !request_mem_region( start, siz, me ) )
			{
				printk(KERN_ERR "%s:  mem region collision at %p\n", me, (void*) start);
				return 1;
			}
			card_hw->shared512k = ioremap( start, siz );
			if ( !card_hw->shared512k )
			{
				printk(KERN_ERR "%s:  ioremap failed\n", me);
				return 1;
			}
			card_hw->shared512k_physaddr = start;

			bars_found++;
			// continue 'for' loop
			continue;
		}
	}
	if ( bars_found < 3 )
	{
		printk(KERN_ERR "%s:  failed to find all needed BARs\n", me);
		return 1;
	}

	return 0;
}
//****************************************************************************


//****************************************************************************
// _oms_attach_irq()
//
//   Note:  spinlock oms_biglock is LOCKED on entry.
int _oms_attach_irq( struct pci_dev* pdev,
										 struct oms_card* card )
{
	int    rv;
	struct list_head* ptr;

	ASSERT_LOCKED();
	
	// see if we've already requested this IRQ
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		struct oms_card* a_card = list_entry( ptr, struct oms_card, listhead );
		if ( a_card->irq == pdev->irq )
		{
			return 0;
		}
	}

	if ( (rv = request_irq(pdev->irq, &_oms_irq_handler, SA_SHIRQ, me, card)) != 0) 
	{
		printk(KERN_ERR "%s:  request_irq failed on irq %d:  %d\n", me, pdev->irq, rv);
		return -1;
	}

	// now we can set the irq number in the card struct
	card->irq = pdev->irq;

	// initialize card's interrupt enable register
  writel( cpu_to_le32(INTERRUPT_MASK), card_hw->obimr );

	// clear the message semaphore and reset the response buffer
	writel( 0, card_hw->semaphore );
	writel( readl( card_hw->response_ip ), card_hw->response_pp );

	return 0;
}
//****************************************************************************


//****************************************************************************
// _oms_detach_irq()
//
//   Note:  spinlock oms_biglock is LOCKED on entry.
void _oms_detach_irq( struct oms_card* card )
{
	ASSERT_LOCKED();
	
	// free the irq
	(void)free_irq( card->irq, card );
}
//****************************************************************************


//****************************************************************************
// _oms_release_resources() --- release all resources used by one card
//
//  Note:  spinlock oms_biglock is LOCKED on entry.
void _oms_release_resources( struct oms_card* card )
{
	ASSERT_LOCKED();

	if ( card->hardware )
	{
		if ( card_hw->shared512k ) 
		{
			iounmap( card_hw->shared512k );
		}
		if ( card_hw->shared64k ) 
		{
			iounmap( card_hw->shared64k );
		}
		if ( card_hw->regbase_addr ) 
		{
			iounmap( (void*)card_hw->regbase_addr );
		}
		if ( card_hw->shared512k_physaddr ) 
		{
			release_mem_region( card_hw->shared512k_physaddr, (512 * 1024) );
		}
		if ( card_hw->shared64k_physaddr ) 
		{
			release_mem_region( card_hw->shared64k_physaddr, (64 * 1024) );
		}
		if ( card_hw->regbase_physaddr ) 
		{
			release_mem_region( card_hw->regbase_physaddr, (4 * 1024) );
		}
		kfree( card->hardware );
	}
	if ( card->irq >= 0 )
		_oms_detach_irq( card );
}
//****************************************************************************


//****************************************************************************
// _oms_fop_read() --- valid ONLY if card->comm_mode is INTERACTIVE_MODE.
//
// NON-blocking read.  If there's nothing to read, returns 0 immediately.
//     Use select(2) or poll(2) to avoid busy looping.  Each time we're
//     called we grab ALL data from the card into a temporary buffer, and
//     copy out what will fit in the user's buffer.  Any residue in
//     temporary buffer is silently discarded.  This is semi-okay, as
//     INTERACTIVE_MODE is intended only for use by OMS user programs,
//     which can be developed to allocate a large enough buffer.
int  _oms_fop_read( struct oms_card* card,
                    char buffer[],
                    int maxsize )
{
	int    rv = 0;
	
	ASSERT_LOCKED();
	
	if ( card->response_ready )
	{
		while ( (rv < maxsize) && (_oms_recv_space_used( card ) > 0) )
		{
			buffer[rv++] = card->recv_buffer[card->recv_rem_ptr];
			card->recv_buffer[card->recv_rem_ptr] = 0;
			if ( ++card->recv_rem_ptr >= INTERACTIVE_BUFFER_LEN )
				card->recv_rem_ptr = 0;
		}
		card->response_ready = FALSE;
	}
	return rv;
}
//****************************************************************************


//****************************************************************************
// _oms_fop_write --- valid only in INTERACTIVE_MODE (returns -ENODEV if not).
//  NON-blocking write.  If there's room in the command buffer, the
//  user data is put there.  If not, we immediately return 0.  Note that
//  one can NEVER successfully write data whose length is greater than
//  the size of the card's command buffer!  (We return -EINVAL to so indicate.)
//
//  In order to avoid a user-program consuming CPU cycles by busily
//  re-issuing a write(2) syscall that return 0, we delay 2 jiffies
//  before we return 0.
int  _oms_fop_write( struct oms_card* card,
                     char buffer[],
                     int maxsize )
{
	int     rv = 0;
	int			size = maxsize;

	ASSERT_LOCKED();

	if ( maxsize <= _oms_xmit_space_avail( card ) )
	{
		// put data into card command buffer
		u32 buf_idx = be32_to_cpu( readl( card_hw->command_ip ) );
		unsigned char* src = buffer;
		volatile unsigned char* dest = card_hw->command_buf;
		
		while ( size--  > 0 )
		{
			dest[buf_idx++] = *src++;
			if ( buf_idx >= CARD_COMMAND_BUFFER_LEN )
				buf_idx = 0;
		}
		writel( cpu_to_be32( buf_idx ), card_hw->command_ip );
		rv = maxsize;
	}

	return rv;
}
//****************************************************************************

	
//****************************************************************************
// We don't buffer outgoing data.  This data is written directly to the card.
//
int  _oms_xmit_space_avail( struct oms_card* card )
{
	u32 get_idx, put_idx;
	
	get_idx = be32_to_cpu( readl( card_hw->command_pp ) );
	put_idx = be32_to_cpu( readl( card_hw->command_ip ) );

	if ( get_idx == put_idx )
		return CARD_COMMAND_BUFFER_LEN - 1;
	else if ( put_idx > get_idx )
		return CARD_COMMAND_BUFFER_LEN - (put_idx - get_idx) - 1;
	else
		return (get_idx - put_idx) - 1;
}
//****************************************************************************


//****************************************************************************
// Incoming data is buffered in the card object's buffer space. An interrupt
// is generated by the card when a complete response is ready and the ISR 
// moves the new data into the local receive buffer.
//
int  _oms_recv_space_used( struct oms_card* card )
{
	if ( card->recv_ins_ptr >= card->recv_rem_ptr )
		return (card->recv_ins_ptr - card->recv_rem_ptr);
	return INTERACTIVE_BUFFER_LEN - (card->recv_rem_ptr - card->recv_ins_ptr - 1);
}
//****************************************************************************


//****************************************************************************
// _oms_irq_handler()
irqreturn_t _oms_irq_handler( int irq, void* dev, struct pt_regs* regs )
{
	irqreturn_t        retval;
	struct list_head*  pos;
	struct oms_card*   card;
	u32                interrupt_mask;
	u32					       curr_status;

	ASSERT_UNLOCKED();
	_LOCK_(flags);

	retval = IRQ_NONE;

	for ( pos = oms_cardlist.next; pos != &oms_cardlist; pos = pos->next )
	{
		card = list_entry( pos, struct oms_card, listhead );

		// disable interrupts, retain present config to reenable later
		interrupt_mask = le32_to_cpu( readl( card_hw->obimr ) );
		writel( 0, card_hw->obimr );

		// read the interrupt status register.  This is LITTLE_ENDIAN
		curr_status = le32_to_cpu( readl( card_hw->obisr ) );

//		if ( curr_status & REG_OBISR_DR )
		{
			u32 data, data2;
			u8 new_status = 0;

			// read the status data reg (it's LITTLE_ENDIAN)
			data = le32_to_cpu( readl( card_hw->obdr ) );
			// construct an old-style status word
			if ( (data & OBDR_CMDERR) > 0 )
			{
				new_status |= CMDERR_BIT;
			}
			if ( (data & OBDR_SLIP_MASK) > 0 )
			{
				new_status |= SLIP_BIT;
				// Save the slip bits
				card_hw->slip_flags |= (u8)((data & OBDR_SLIP_MASK) >> OBDR_SLIP_SHIFT);
			}
			if ( (data & OBDR_LIMIT_MASK) > 0 )
			{
				new_status |= OVRT_BIT;
				// Save the limit bits
				data2 = le32_to_cpu( readl( card_hw->limit ) );
				card_hw->pos_limit_flags |= (u8)((data2 & MEM_LIMIT_POS_MASK) >> MEM_LIMIT_POS_SHIFT);
				card_hw->neg_limit_flags |= (u8)((data2 & MEM_LIMIT_NEG_MASK) >> MEM_LIMIT_NEG_SHIFT);
			}
			if ( (data & OBDR_DONE_MASK) > 0 )
			{
				new_status |= DONE_BIT;
				// Save the done bits
				card_hw->done_flags |= (u8)((data & OBDR_DONE_MASK) >> OBDR_DONE_SHIFT);
			}

			// clear the status data register
			writel( cpu_to_le32( data ), card_hw->obdr );

			// One more bit... Check for init condition
			data = le32_to_cpu( readl( card_hw->firmware ) );
			if ( (data & MEM_FW_POWERUP) == MEM_FW_POWERUP || (data & MEM_FW_INITING) == MEM_FW_INITING )
				new_status |= INIT_BIT;
				
			// preserve status bits in our card struct
			card_hw->status_flags |= new_status;
			// dismiss the interrupt condition
			writel( cpu_to_le32( REG_OBISR_DR ), card_hw->obisr );
			retval = IRQ_HANDLED;
		}
		
		if ( curr_status & REG_OBISR_MR0 ) 
		{
			if ( (be32_to_cpu( readl( card_hw->obmr0 ) ) & OBMR0_RESPONSE) == OBMR0_RESPONSE )
			{
				_oms_irq_handle_ibf( card );
				// clear card's text semaphore
				writel( 0, card_hw->semaphore );
			}
			// dismiss the interrupt condition
			writel( cpu_to_le32( REG_OBISR_MR0 ), card_hw->obisr );
			retval = IRQ_HANDLED;
		}
    
		if ( curr_status & REG_OBISR_MR1 )
		{
			if ( be32_to_cpu( readl( card_hw->obmr1 ) ) == OBMR1_POSDATA )
			{
				_oms_irq_handle_pos_data( card );
			}
			else
			{
				printk(KERN_INFO "%s:  bad MSG_REG_1 irq\n", me);
			}
			// dismiss the interrupt condition
			writel( cpu_to_le32( REG_OBISR_MR1 ), card_hw->obisr );
			retval = IRQ_HANDLED;
		}

		// reenable previously enabled interrupts
		writel( interrupt_mask, card_hw->obimr );
	}
	
	UN_LOCK_(flags);

	return retval;
}
//****************************************************************************


//****************************************************************************
// _oms_irq_handle_ibf() --- called from _oms_irq_handler() when
//   the card indicates that there is a TEXT_RESPONSE.  Determine from
//   the state of the card and the ioctl op queue what to do with the
//   data.
//
//  Note:  the 'biglock' spinlock is locked.
void _oms_irq_handle_ibf( struct oms_card* card )
{
	struct list_head*     pos;
	struct omsdrv_ioctl*  pioctl = NULL;
	u32                   obmr0;
	u32                   index;
	u32                   len;
	char                  tempbuf[CARD_RESPONSE_BUFFER_LEN+1];
	char                  resp_char;
	int                   i = 0;
	
	ASSERT_LOCKED();

	// Collect the response from the card
	obmr0 = be32_to_cpu( readl( card_hw->obmr0 ) );
	index = ((obmr0 & OBMR0_START_MASK) >> OBMR0_START_SHIFT);
	len = ((obmr0 & OBMR0_LEN_MASK) >> OBMR0_LEN_SHIFT);
	for ( ; len > 0; len-- )
	{
		resp_char = readb( card_hw->response_buf + index );
		if ( resp_char != LF && resp_char != CR )
		{
			tempbuf[i++] = resp_char;
		}
		if ( ++index >= CARD_RESPONSE_BUFFER_LEN )
			index -= CARD_RESPONSE_BUFFER_LEN;
	}
	tempbuf[i] = 0;
	
	// Decide what to do with the collected response
	if ( card->comm_mode == INTERACTIVE_MODE )
	{
		i = 0;
		while ( tempbuf[i] != 0 )
		{
			card->recv_buffer[card->recv_ins_ptr] = tempbuf[i++];
			if ( ++card->recv_ins_ptr >= INTERACTIVE_BUFFER_LEN )
				card->recv_ins_ptr = 0;
		}
		card->lf_count = 0;
		card->response_ready = TRUE;
		wake_up( &card->poll_waitq );
	}
	else
	{
		// find the ioctl associated with the last command sent
		if ( list_empty( &card->ioctl_list ) ) 
		{
			// no ioctls waiting
			return;
		}
		pos = card->ioctl_list.next;
		pioctl = list_entry( pos, struct omsdrv_ioctl, listhead );
		if ( pioctl->response_expected == FALSE ) 
		{
			if ( pioctl->buffers.command && 
					 pioctl->p_command < pioctl->buffers.command_len )
			{
				if ( DEBUGLEVEL > 2 )
					printk( KERN_INFO "%s: IBF IRQ received but still transmitting\n", me );
				// still transmitting, just exit
				return;
			}
			
			printk(KERN_ERR "%s: irq:  unexpected response from card\n", me);
			if ( pioctl->task_asleep ) 
			{
				wake_up( &pioctl->waitq );
			}
			else 
			{
				oms_release_ioctl( pioctl );
			}
			return;
		}

		// ioctl found.  store the response and wake up the ioctl
		i = 0;
		while ( tempbuf[i] != 0 )
		{
			pioctl->buffers.response[pioctl->p_response++] = tempbuf[i++];
		}

		pioctl->lf_count = 0;
		pioctl->card->response_ready = 1;
					
		// let the task complete the ioctl operation
		if ( pioctl->task_asleep ) 
		{
			wake_up( &pioctl->waitq );
		} 
		else 
		{
			printk(KERN_ERR "%s:  irq:  task not asleep for response\n", me);
			oms_release_ioctl( pioctl );
		}
	}
}
//****************************************************************************


//****************************************************************************
// _oms_irq_handle_pos_data() --- called from _oms_irq_handler()
//
//  Note:  the 'biglock' spinlock is locked.
void _oms_irq_handle_pos_data( struct oms_card* card )
{
	struct list_head*     pos;
	struct omsdrv_ioctl*  pioctl = NULL;
  u32*                  pbuffer;
  u32*                  pmemory;
        int                   i;

	ASSERT_LOCKED();

	if ( list_empty( &card->ioctl_list ) ) 
	{
		// no ioctls waiting
		return;
	}
	pos = card->ioctl_list.next;
	pioctl = list_entry( pos, struct omsdrv_ioctl, listhead );
	if ( pioctl == NULL || pioctl->response_expected == FALSE ||
			 pioctl->cmd != IOCTL_OMS_READ_AXIS_POSITIONS ) 
	{
		// not expecting position data
		return;
	}

	pbuffer = (u32*)(pioctl->buffers.response);
  pmemory = (u32*)(card_hw->shared64k + MEM_CAP_MOTOR);
	for ( i = 0; i < 8; i++ )
	{
		pbuffer[i] = be32_to_cpu( readl( &pmemory[i] ) );
	}
	
  pmemory = (u32*)(card_hw->shared64k + MEM_CAP_ENCODER);
	for ( i = 0; i < 8; i++ )
	{
		pbuffer[i+8] = be32_to_cpu( readl( &pmemory[i] ) );
	}
	
	pioctl->buffers.response_len = 16 * sizeof(u32);
	// let the task complete the ioctl operation
	if ( pioctl->task_asleep ) 
	{
		wake_up( &pioctl->waitq );
	} 
	else 
	{
		printk(KERN_ERR "%s:  irq:  task not asleep for response\n", me);
		oms_release_ioctl( pioctl );
	}
}
//****************************************************************************


//****************************************************************************
// Support card-specific IOCTL operations
//
int  _oms_fop_ioctl( struct inode* pinode, 
                     struct file* pfile,
                     unsigned int cmd,
                     unsigned long userarg )
{
/*
	int                   minor;
	struct oms_card*      card;
	struct list_head*     ptr;
	struct omsdrv_ioctl*  pioctl;
	int                   rv;
	void*                 arg = (void*)userarg;

	ASSERT_UNLOCKED();

	// find oms card by minor	
	minor = MINOR( pfile->f_dentry->d_inode->i_rdev );
	if ( minor < 0 || minor >= MAX_CARDS )
		return -ENODEV;
	for ( ptr = oms_cardlist.next; ptr != &oms_cardlist; ptr = ptr->next )
	{
		card = list_entry( ptr, struct oms_card, listhead );
		if ( card->minor == minor )
			break;
	}
	if ( ptr == &oms_cardlist ) 
	{
		return -ENODEV;
	}
*/

	// execute requested ioctl function
	switch (cmd) 
	{
		default:
			return -EINVAL;
	}
	return 0;
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_read_status_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// read the status_flags.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 1, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != sizeof(card_hw->status_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;
	memset( pioctl->buffers.response, 0, sizeof(card_hw->status_flags) );  // pre-zero

	_LOCK_(flags);
	memcpy( pioctl->buffers.response, &card_hw->status_flags, sizeof(card_hw->status_flags) );
	UN_LOCK_(flags);

	if ( pioctl->buffers.response[0] != 0 )
	{
		pioctl->p_response = sizeof(card_hw->status_flags);
		pioctl->response_expected = TRUE; // fake out our copy_out func
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	}
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_status_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// clear status flag bits selected by 1's in the command
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.command_len != sizeof(card_hw->status_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	switch ( sizeof(card_hw->status_flags) )
	{
		case 1:
		{
			u8 mask = *((u8*)(pioctl->buffers.command));
			card_hw->status_flags &= ~mask;
			break;
		}
		case 2:
		{
			u16 mask = *((u16*)(pioctl->buffers.command));
			card_hw->status_flags &= ~mask;
			break;
		}
		case 4:
		{
			u32 mask = *((u32*)(pioctl->buffers.command));
			card_hw->status_flags &= ~mask;
			break;
		}
	}
	UN_LOCK_(flags);

	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_done_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// read the done_flags.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 1, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != sizeof(card_hw->done_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;
	memset( pioctl->buffers.response, 0, sizeof(card_hw->done_flags) );

	_LOCK_(flags);
	memcpy( pioctl->buffers.response, &card_hw->done_flags, sizeof(card_hw->done_flags) );
	UN_LOCK_(flags);

	if ( pioctl->buffers.response[0] != 0 )
	{
		pioctl->p_response = sizeof(card_hw->done_flags);
		pioctl->response_expected = TRUE; // fake out our copy_out func
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	}
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_done_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// clear flag bits selected by 1's in the command
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.command_len != sizeof(card_hw->done_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	switch ( sizeof(card_hw->done_flags) )
	{
		case 1:
		{
			u8 mask = *((u8*)(pioctl->buffers.command));
			card_hw->done_flags &= ~mask;
			break;
		}
		case 2:
		{
			u16 mask = *((u16*)(pioctl->buffers.command));
			card_hw->done_flags &= ~mask;
			break;
		}
		case 4:
		{
			u32 mask = *((u32*)(pioctl->buffers.command));
			card_hw->done_flags &= ~mask;
			break;
		}
	}
	UN_LOCK_(flags);

	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_pos_limit_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// read the flags.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 1, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != sizeof(card_hw->pos_limit_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;
	memset( pioctl->buffers.response, 0, sizeof(card_hw->pos_limit_flags) );

	_LOCK_(flags);
	memcpy( pioctl->buffers.response, &card_hw->pos_limit_flags, sizeof(card_hw->pos_limit_flags) );
	UN_LOCK_(flags);

	if ( pioctl->buffers.response[0] != 0 )
	{
		pioctl->p_response = sizeof(card_hw->pos_limit_flags);
		pioctl->response_expected = TRUE; // fake out our copy_out func
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	}
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_pos_limit_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// clear flag bits selected by 1's in the command
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.command_len != sizeof(card_hw->pos_limit_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	switch ( sizeof(card_hw->pos_limit_flags) )
	{
		case 1:
		{
			u8 mask = *((u8*)(pioctl->buffers.command));
			card_hw->pos_limit_flags &= ~mask;
			break;
		}
		case 2:
		{
			u16 mask = *((u16*)(pioctl->buffers.command));
			card_hw->pos_limit_flags &= ~mask;
			break;
		}
		case 4:
		{
			u32 mask = *((u32*)(pioctl->buffers.command));
			card_hw->pos_limit_flags &= ~mask;
			break;
		}
	}
	UN_LOCK_(flags);

	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_neg_limit_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// read the flags.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 1, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != sizeof(card_hw->neg_limit_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;
	memset( pioctl->buffers.response, 0, sizeof(card_hw->neg_limit_flags) );

	_LOCK_(flags);
	memcpy( pioctl->buffers.response, &card_hw->neg_limit_flags, sizeof(card_hw->neg_limit_flags) );
	UN_LOCK_(flags);

	if ( pioctl->buffers.response[0] != 0 )
	{
		pioctl->p_response = sizeof(card_hw->neg_limit_flags);
		pioctl->response_expected = TRUE; // fake out our copy_out func
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	}
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_neg_limit_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// clear flag bits selected by 1's in the command
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.command_len != sizeof(card_hw->neg_limit_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	switch ( sizeof(card_hw->neg_limit_flags) )
	{
		case 1:
		{
			u8 mask = *((u8*)(pioctl->buffers.command));
			card_hw->neg_limit_flags &= ~mask;
			break;
		}
		case 2:
		{
			u16 mask = *((u16*)(pioctl->buffers.command));
			card_hw->neg_limit_flags &= ~mask;
			break;
		}
		case 4:
		{
			u32 mask = *((u32*)(pioctl->buffers.command));
			card_hw->neg_limit_flags &= ~mask;
			break;
		}
	}
	UN_LOCK_(flags);

	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_slip_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// read the flags.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 1, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != sizeof(card_hw->slip_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;
	memset( pioctl->buffers.response, 0, sizeof(card_hw->slip_flags) );

	_LOCK_(flags);
	memcpy( pioctl->buffers.response, &card_hw->slip_flags, sizeof(card_hw->slip_flags) );
	UN_LOCK_(flags);

	if ( pioctl->buffers.response[0] != 0 )
	{
		pioctl->p_response = sizeof(card_hw->slip_flags);
		pioctl->response_expected = TRUE; // fake out our copy_out func
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	}
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_clear_slip_flags( struct oms_card* card, void* arg )
{
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	// clear flag bits selected by 1's in the command
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 1, 0, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.command_len != sizeof(card_hw->slip_flags) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}
	//pioctl->cmd = cmd;

	_LOCK_(flags);
	switch ( sizeof(card_hw->slip_flags) )
	{
		case 1:
		{
			u8 mask = *((u8*)(pioctl->buffers.command));
			card_hw->slip_flags &= ~mask;
			break;
		}
		case 2:
		{
			u16 mask = *((u16*)(pioctl->buffers.command));
			card_hw->slip_flags &= ~mask;
			break;
		}
		case 4:
		{
			u32 mask = *((u32*)(pioctl->buffers.command));
			card_hw->slip_flags &= ~mask;
			break;
		}
	}
	UN_LOCK_(flags);

	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************

		
//****************************************************************************
int  _oms_ioctl_read_shared_mem( struct oms_card* card, void* arg )
{
	u32*									src;
	u32*									dest;
	u32										count, word, i, offst;
	struct omsdrv_ioctl*  pioctl;
	int                   rv;

	if ( card->comm_mode == INTERACTIVE_MODE )
		return -ENODEV;

	// read shared memory.  The starting offset is the
	//   single u32 of the command string, the number
	//   of u32's to read is given by the response_len.
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 2, 2, arg )) != 0 )
		return rv;  // pioctl already released

	// verify that command_len == sizeof(u32), i.e., one word
	//   also verify a non-zero response length
	if ( pioctl->buffers.command_len != sizeof(u32) ||	
			 pioctl->buffers.response_len == 0) 
	{
		oms_release_ioctl( pioctl );
			return -EINVAL;
	}

	_LOCK_(flags);
	// copy data from the shared memory address to pioctl
	offst = *((u32*)pioctl->buffers.command);
	src = (u32*)(card_hw->shared64k + offst);
	dest = (u32*)(pioctl->buffers.response);
	count = pioctl->buffers.response_len / sizeof(u32);
	// make sure src addr is 32-bit word aligned
	if ( (unsigned long)src & 0x03 ) 
	{
		UN_LOCK_(flags);
		oms_release_ioctl( pioctl );
		return (-EINVAL);
	}
	for ( i = 0; i < count; i++, src++ ) 
	{
		// range check
		if ( (unsigned long)src >=
						(((unsigned long)(card_hw->shared64k)) +
						(64 * 1024)) ) 
		{
			UN_LOCK_(flags);
			oms_release_ioctl( pioctl );
			return (-EINVAL);
		}
		word = be32_to_cpu( readl( src ) );
		*dest++ = word;
	}
	UN_LOCK_(flags);

	pioctl->p_response = pioctl->buffers.response_len;
	pioctl->response_expected = TRUE;  // fake out copy_out function
	rv = oms_copy_out_ioctl_response( pioctl, arg );
	oms_release_ioctl( pioctl );
	return (rv);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_write_shared_mem( struct oms_card* card, void* arg )
{
	u32*									src;
	u32*									dest;
	u32										count, word, i;
	struct omsdrv_ioctl*  pioctl;
	int                   rv;
			
	// write words (u32's) to shared memory.  The first
	//  word of the command string is the offset into
	//  shared memory.  The second word is the word count,
	//  and the third (and successive) words are the data
	//  to write in big-endian format.
	//
	//  This one is easy to get wrong in your user program.
	//  Make sure that command_len is:
	//             (2 * sizeof(__u32)) + sizeof(data)
	//  And that the 2 words of info (offset and WORD count)
	//  are contiguous with the data.
	//
	//  As an aid, we return -ENXIO if the command_len
	//  doesn't jibe with the word-count.
	if ( card->comm_mode == INTERACTIVE_MODE )
		return -ENODEV;
	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 2, 0, arg )) != 0 )
		return rv;  // pioctl already released
	if ( pioctl->buffers.command_len < (3 * sizeof(u16)) ||
				pioctl->buffers.response_len != 0) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}

	_LOCK_(flags);
	// copy the data into the shared memory
	src = (u32*)pioctl->buffers.command;
	dest = (u32*)(card_hw->shared64k + *src++);
	count = *src++;  // must be a WORD count!!

	// make sure command_len jibes with the count
	if ( pioctl->buffers.command_len != (count + 2) * sizeof(u32) ) 
	{
		UN_LOCK_(flags);
		oms_release_ioctl( pioctl );
		return (-ENXIO);
	}

	// make sure that dest addr is word-aligned
	if ((unsigned long) dest & 0x03) 
	{
		UN_LOCK_(flags);
		oms_release_ioctl( pioctl );
		return (-EINVAL);
	}

	for ( i = 0; i < count; i++, dest++ ) 
	{
		word = *src++;
		// range check
		if ((unsigned long) dest >=
				(((unsigned long)(card_hw->shared64k)) +
				(64 * 1024))) 
		{
			UN_LOCK_(flags);
			oms_release_ioctl( pioctl );
			return (-EINVAL);
		}
		writel( cpu_to_be32( word ), dest );
	}
	UN_LOCK_(flags);
	oms_release_ioctl( pioctl );
	return (0);
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_read_axis_positions( struct oms_card* card, void* arg )
{		
	struct omsdrv_ioctl*  pioctl;
	int                   rv;
	
//	if ( card->comm_mode == INTERACTIVE_MODE )
//		return -ENODEV;

	pioctl = oms_new_ioctl( card );
	if ( (rv = oms_fetch_user_ioctl_struct( pioctl, 0, 4, arg )) != 0 )
		return rv;  // ios already released
	if ( pioctl->buffers.response_len != (16 * sizeof(u32)) ) 
	{
		oms_release_ioctl( pioctl );
		return -EINVAL;
	}

	// tickle the mailbox to get updated position info
  writel( cpu_to_be32(MEM_CAP_REQUEST), card_hw->mailbox );

	pioctl->response_expected = TRUE;
	pioctl->cmd = IOCTL_OMS_READ_AXIS_POSITIONS;

	_LOCK_(flags);
	tasklet_schedule( &oms_tasklet );
	rv = oms_enqueue_ioctl_and_sleep( pioctl, &flags );
	UN_LOCK_(flags);

	pioctl->p_response = pioctl->buffers.response_len;
	rv = oms_copy_out_ioctl_response( pioctl, arg );

	if (rv == 0)
		rv = oms_copy_out_ioctl_response( pioctl, arg );
	oms_release_ioctl( pioctl );
	return rv;
}
//****************************************************************************


//****************************************************************************
int  _oms_ioctl_reinit( struct oms_card* card, void* arg )
{
/*
	// Rewrite the interrupt control register
	_LOCK_(flags);
	*card_hw->eg_control = INTERRUPT_MASK;
	UN_LOCK_(flags);
*/
	return 0;
}
//****************************************************************************


//****************************************************************************
void _oms_start_ioctl( struct oms_card* card )
{
/*
	ASSERT_LOCKED();

	// enable TBE interrupts
	*card_hw->reg_control = (*card_hw->reg_control & ~TBE_BIT);
	*card_hw->reg_control = (*card_hw->reg_control | TBE_BIT);
*/
}
//****************************************************************************


